#!/usr/bin/python
# Copyright (C) 2022 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import shutil

from breezy.revision import NULL_REVISION
from breezy.tests import TestCaseWithTransport

from silver_platter.workspace import Workspace


class TestWorkspace(TestCaseWithTransport):

    def test_nascent(self):
        tree = self.make_branch_and_tree('origin')
        with Workspace(tree.branch, dir=self.test_dir) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit('A change')
            self.assertEqual(ws.path, os.path.join(ws.local_tree.basedir, '.'))
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())
            self.assertTrue(ws.any_branch_changes())
            self.assertEqual(
                [('', NULL_REVISION, ws.local_tree.last_revision())],
                ws.result_branches())

    def test_without_main(self):
        with Workspace(None, dir=self.test_dir) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit('A change')
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())
            self.assertTrue(ws.any_branch_changes())
            self.assertEqual(
                [('', NULL_REVISION, ws.local_tree.last_revision())],
                ws.result_branches())

    def test_basic(self):
        tree = self.make_branch_and_tree('origin')
        revid1 = tree.commit('first commit')
        with Workspace(tree.branch, dir=self.test_dir) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit('A change')
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())
            self.assertTrue(ws.any_branch_changes())
            self.assertEqual(
                [('', revid1, ws.local_tree.last_revision())],
                ws.result_branches())

    def test_cached_branch_up_to_date(self):
        tree = self.make_branch_and_tree('origin')
        revid1 = tree.commit('first commit')
        cached = tree.branch.controldir.sprout('cached')
        with Workspace(tree.branch, cached_branch=cached.open_branch(),
                       dir=self.test_dir) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertFalse(ws.changes_since_base())
            self.assertEqual(ws.local_tree.last_revision(), revid1)

    def test_cached_branch_out_of_date(self):
        tree = self.make_branch_and_tree('origin')
        tree.commit('first commit')
        cached = tree.branch.controldir.sprout('cached')
        revid2 = tree.commit('first commit')
        with Workspace(tree.branch, cached_branch=cached.open_branch(),
                       dir=self.test_dir) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertFalse(ws.changes_since_base())
            self.assertEqual(ws.local_tree.last_revision(), revid2)

    def commit_on_colo(self, controldir, name, message):
        colo_branch = controldir.create_branch('colo')
        colo_checkout = colo_branch.create_checkout(name)
        try:
            return colo_checkout.commit(message)
        finally:
            shutil.rmtree(name)

    def test_colocated(self):
        tree = self.make_branch_and_tree('origin')
        revid1 = tree.commit('main')
        colo_revid1 = self.commit_on_colo(
            tree.branch.controldir, 'colo', 'Another')
        self.assertEqual(tree.branch.last_revision(), revid1)
        with Workspace(
                tree.branch, dir=self.test_dir,
                additional_colocated_branches=['colo']) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit('A change')
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())
            self.assertTrue(ws.any_branch_changes())
            self.assertEqual(
                [('', revid1, ws.local_tree.last_revision()),
                 ('colo', colo_revid1, colo_revid1)],
                ws.result_branches())

    def test_resume_continue(self):
        tree = self.make_branch_and_tree('origin')
        revid1 = tree.commit('first commit')
        resume = tree.branch.controldir.sprout('resume')
        resume_tree = resume.open_workingtree()
        resume_revid1 = resume_tree.commit('resume')
        with Workspace(tree.branch, resume_branch=resume_tree.branch,
                       dir=self.test_dir) as ws:
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.any_branch_changes())
            self.assertFalse(ws.refreshed)
            self.assertFalse(ws.changes_since_base())
            self.assertEqual(ws.local_tree.last_revision(), resume_revid1)
            self.assertEqual([
                ('', revid1, resume_revid1)], ws.result_branches())

    def test_resume_discard(self):
        tree = self.make_branch_and_tree('origin')
        tree.commit('first commit')
        resume = tree.branch.controldir.sprout('resume')
        revid2 = tree.commit('second commit')
        resume_tree = resume.open_workingtree()
        resume_tree.commit('resume')
        with Workspace(tree.branch, resume_branch=resume_tree.branch,
                       dir=self.test_dir) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertTrue(ws.refreshed)
            self.assertFalse(ws.changes_since_base())
            self.assertEqual(ws.local_tree.last_revision(), revid2)
            self.assertEqual([('', revid2, revid2)], ws.result_branches())

    def test_resume_continue_with_unchanged_colocated(self):
        tree = self.make_branch_and_tree('origin')
        revid1 = tree.commit('first commit')
        colo_revid1 = self.commit_on_colo(
            tree.branch.controldir, 'colo', 'First colo')
        resume = tree.branch.controldir.sprout('resume')
        resume_tree = resume.open_workingtree()
        resume_revid1 = resume_tree.commit('resume')
        with Workspace(tree.branch, resume_branch=resume_tree.branch,
                       dir=self.test_dir,
                       additional_colocated_branches=['colo']) as ws:
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.any_branch_changes())
            self.assertFalse(ws.refreshed)
            self.assertFalse(ws.changes_since_base())
            self.assertEqual(ws.local_tree.last_revision(), resume_revid1)
            self.assertEqual([
                ('', revid1, resume_revid1),
                ('colo', colo_revid1, colo_revid1),
            ], ws.result_branches())

    def test_resume_discard_with_unchanged_colocated(self):
        tree = self.make_branch_and_tree('origin')
        tree.commit('first commit')
        colo_revid1 = self.commit_on_colo(
            tree.branch.controldir, 'colo', 'First colo')
        resume = tree.branch.controldir.sprout('resume')
        self.commit_on_colo(resume, 'colo', 'First colo on resume')
        revid2 = tree.commit('second commit')
        resume_tree = resume.open_workingtree()
        resume_tree.commit('resume')
        with Workspace(tree.branch, resume_branch=resume_tree.branch,
                       dir=self.test_dir,
                       additional_colocated_branches=['colo']) as ws:
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.any_branch_changes())
            self.assertTrue(ws.refreshed)
            self.assertFalse(ws.changes_since_base())
            self.assertEqual(ws.local_tree.last_revision(), revid2)
            self.assertEqual([
                ('', revid2, revid2),
                ('colo', colo_revid1, colo_revid1),
            ], ws.result_branches())



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
import subprocess
import sys
from typing import List, Optional

from breezy.workingtree import WorkingTree
from breezy.workspace import check_clean_tree, reset_tree
from . import _svp_rs


ScriptMadeNoChanges = _svp_rs.ScriptMadeNoChanges
ScriptFailed = _svp_rs.ScriptFailed
ScriptNotFound = _svp_rs.ScriptNotFound
DetailedFailure = _svp_rs.DetailedFailure
ResultFileFormatError = _svp_rs.ResultFileFormatError
CommandResult = _svp_rs.CommandResult
script_runner = _svp_rs.script_runner


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command", help="Path to script to run.", type=str,
        nargs='?')
    parser.add_argument(
        "--diff", action="store_true", help="Show diff of generated changes."
    )
    parser.add_argument(
        "--commit-pending",
        help="Commit pending changes after script.",
        choices=["yes", "no", "auto"],
        default=None,
        type=str,
    )
    parser.add_argument(
        "--verify-command", type=str, help="Command to run to verify changes."
    )
    parser.add_argument(
        "--recipe", type=str, help="Recipe to use.")
    args = parser.parse_args(argv)

    if args.recipe:
        from .recipe import Recipe
        recipe = Recipe.from_path(args.recipe)
    else:
        recipe = None

    if args.commit_pending:
        commit_pending = {
            "auto": None, "yes": True, "no": False}[args.commit_pending]
    elif recipe:
        commit_pending = recipe.commit_pending
    else:
        commit_pending = None

    if args.command:
        command = args.command
    elif recipe.command:
        command = recipe.command
    else:
        parser.error('No command specified.')

    local_tree, subpath = WorkingTree.open_containing('.')

    check_clean_tree(local_tree)

    try:
        result = script_runner(
            local_tree, script=command, commit_pending=commit_pending,
            subpath=subpath)

        if result.description:
            logging.info('Succeeded: %s', result.description)

        if args.verify_command:
            try:
                subprocess.check_call(
                    args.verify_command, shell=True,
                    cwd=local_tree.abspath(subpath)
                )
            except subprocess.CalledProcessError:
                parser.error("Verify command failed.")
    except Exception:
        reset_tree(local_tree, subpath=subpath)
        raise

    if args.diff:
        from breezy.diff import show_diff_trees
        old_tree = local_tree.revision_tree(result.old_revision)
        new_tree = local_tree.revision_tree(result.new_revision)
        show_diff_trees(
            old_tree,
            new_tree,
            sys.stdout.buffer,
            old_label='old/',
            new_label='new/')
    return 0



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest


def test_suite():
    names = [
        "apply",
        "candidates",
        "debian",
        "probers",
        "proposal",
        "publish",
        "recipe",
        "run",
        "utils",
        "workspace",
    ]
    module_names = [__name__ + ".test_" + name for name in names]
    loader = unittest.TestLoader()
    return loader.loadTestsFromNames(module_names)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# TODO(jelmer): Imports with side-effects are bad...

import breezy  # noqa: F401
import breezy.bzr  # For bzr support   # noqa: F401
import breezy.git  # For git support   # noqa: F401
import breezy.plugins.github  # For github support  # noqa: F401
import breezy.plugins.gitlab  # For gitlab support  # noqa: F401
import breezy.plugins.launchpad  # For lp: URL support  # noqa: F401

__version__ = (0, 5, 11)
version_string = ".".join(map(str, __version__))



#!/usr/bin/python
# Copyright (C) 2022 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
import os
import shutil
import sys
from contextlib import suppress
from typing import Any, Callable, Dict, List, Optional

import ruamel.yaml
from breezy.branch import Branch
from breezy.diff import show_diff_trees
from breezy.errors import DivergedBranches
from breezy.forge import get_proposal_by_url
from breezy.workingtree import WorkingTree
from ruamel.yaml.scalarstring import LiteralScalarString

from .apply import (
    ScriptFailed,
    ScriptMadeNoChanges,
    ScriptNotFound,
    script_runner,
)
from .candidates import Candidate, CandidateList
from .proposal import (
    ForgeLoginRequired,
    MergeProposal,
    UnsupportedForge,
    enable_tag_pushing,
    get_forge,
)
from .publish import (
    EmptyMergeProposal,
    InsufficientChangesForNewProposal,
    publish_changes,
)
from .recipe import Recipe
from .utils import (
    BranchMissing,
    BranchUnavailable,
    BranchUnsupported,
    full_branch_url,
    open_branch,
)
from .workspace import Workspace


def generate_for_candidate(recipe, basepath, url, *, subpath: str = '',
                           default_mode: Optional[str] = None):
    try:
        main_branch = open_branch(url)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        logging.error("%s: %s", url, e)
        raise

    with Workspace(main_branch, path=basepath) as ws:
        logging.info('Making changes to %s', main_branch.user_url)
        entry = {'url': url}
        if subpath:
            entry['subpath'] = subpath

        try:
            result = script_runner(
                ws.local_tree, recipe.command,
                commit_pending=recipe.commit_pending,
                subpath=subpath)
        except ScriptMadeNoChanges:
            logging.error("Script did not make any changes.")
            return None
        except ScriptFailed:
            logging.error("Script failed to run.")
            return None
        except ScriptNotFound:
            logging.error("Script not found.")
            return None
        else:
            if result.target_branch_url:
                entry['target_branch_url'] = result.target_branch_url
            if result.description:
                entry['description'] = LiteralScalarString(result.description)
            else:
                description = recipe.render_merge_request_description(
                    'markdown', result.context)
                if description:
                    entry['description'] = LiteralScalarString(description)
            commit_message = recipe.render_merge_request_commit_message(
                result.context)
            if commit_message:
                entry['commit-message'] = commit_message
            title = recipe.render_merge_request_title(
                result.context)
            if title:
                entry['title'] = title
            if recipe.mode:
                entry['mode'] = recipe.mode or default_mode
            if recipe.labels:
                entry['labels'] = recipe.labels
            if result.context:
                entry['context'] = recipe.context
            ws.defer_destroy()
        return entry


def generate(
        recipe: Recipe, candidates: List[Candidate], directory: str,
        recipe_path: str):
    with suppress(FileExistsError):
        os.mkdir(directory)

    batch: Dict[str, Any]
    try:
        batch = load_batch_metadata(directory)
    except FileNotFoundError:
        batch = {
            'recipe': recipe_path,
            'name': recipe.name,
        }
        batch['work'] = entries = {}
    else:
        entries = batch['work']

    try:
        for candidate in candidates:
            basename = candidate.name
            if basename is None:
                # TODO(jelmer): Move this logic to Candidate?
                basename = candidate.url.rstrip('/').rsplit('/', 1)[-1]
            name = basename
            # TODO(jelmer): Search by URL rather than by name?
            if name in entries and entries[name]['url'] == candidate.url:
                logging.info(
                    'An entry %s for %s exists, skipping',
                    name, entries[name]['url'])
                continue
            i = 0
            while os.path.exists(os.path.join(directory, name)):
                i += 1
                name = basename + '.%d' % i
            work_path = os.path.join(directory, name)
            try:
                entry = generate_for_candidate(
                    recipe, work_path,
                    candidate.url,
                    subpath=candidate.subpath or '',
                    default_mode=candidate.default_mode)
            except Exception:
                if os.path.exists(work_path):
                    shutil.rmtree(work_path)
                raise
            else:
                if entry:
                    entries[name] = entry
                    save_batch_metadata(directory, batch)
    finally:
        save_batch_metadata(directory, batch)


def drop_batch_entry(directory, name):
    with suppress(FileNotFoundError):
        shutil.rmtree(os.path.join(directory, name))


def save_batch_metadata(directory: str, batch) -> None:
    with open(os.path.join(directory, 'batch.yaml'), 'w') as f:
        ruamel.yaml.round_trip_dump(batch, f)


def load_batch_metadata(directory):
    with open(os.path.join(directory, 'batch.yaml')) as f:
        return ruamel.yaml.round_trip_load(f)


def status(directory, codebase=None):
    batch = load_batch_metadata(directory)
    work = batch.get('work', [])
    if not work:
        logging.error('no work found in %s', directory)
        return 0
    for name, entry in work.items():
        if codebase is not None and name != codebase:
            continue
        if entry.get('proposal-url'):
            proposal = get_proposal_by_url(entry['proposal-url'])
            if proposal.is_merged():
                logging.info('%s: %s was merged', name,
                             entry['proposal-url'])
            elif proposal.is_closed():
                logging.info('%s: %s was closed without being merged',
                             name, entry['proposal-url'])
            else:
                logging.info('%s: %s is still open',
                             name, entry['proposal-url'])
        else:
            logging.info('%s: not published yet', name)


def diff(directory, codebase):
    batch = load_batch_metadata(directory)
    work = batch.get('work', [])
    if not work:
        logging.error('no work found in %s', directory)
        return 0
    entry = work[codebase]
    try:
        main_branch = open_branch(entry['url'])
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        logging.error("%s: %s", entry['url'], e)
        raise

    basepath = os.path.join(directory, codebase)
    local_branch = Branch.open(basepath)
    show_diff_trees(
        local_branch.basis_tree(),
        local_branch.repository.revision_tree(main_branch.last_revision()),
        sys.stdout.buffer,
        old_label='old/',
        new_label='new/',
    )


class UnrelatedBranchExists(Exception):
    """An unrelated branch exists."""


def publish_one(url: str, path: str, batch_name: str, mode: str,
                *,
                subpath: str = '',
                existing_proposal_url: Optional[str] = None,
                labels: Optional[List[str]] = None,
                derived_owner: Optional[str] = None, refresh: bool = False,
                commit_message: Optional[str] = None,
                title: Optional[str] = None,
                description: Optional[str] = None, overwrite: bool = False):
    try:
        main_branch = open_branch(url)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        logging.error("%s: %s", url, e)
        raise

    try:
        forge = get_forge(main_branch)
    except UnsupportedForge as e:
        if mode != "push":
            raise
        # We can't figure out what branch to resume from when there's no forge
        # that can tell us.
        resume_branch = None
        existing_proposal: Optional[MergeProposal] = None
        logging.warn(
            "Unsupported forge (%s), will attempt to push to %s",
            e,
            full_branch_url(main_branch),
        )
    else:
        if existing_proposal_url is not None:
            existing_proposal = forge.get_proposal_by_url(
                existing_proposal_url)
            assert existing_proposal
            resume_branch_url = existing_proposal.get_source_branch_url()
            assert resume_branch_url is not None
            resume_branch = Branch.open(resume_branch_url)
        else:
            existing_proposal = None
            resume_branch = None
    if refresh:
        if resume_branch:
            overwrite = True
        resume_branch = None

    if existing_proposal:
        logging.info('Updating %s', existing_proposal.url)
    else:
        existing_proposal = None

    local_tree = WorkingTree.open(path)

    enable_tag_pushing(local_tree.branch)

    try:
        publish_result = publish_changes(
            local_tree.branch,
            main_branch,
            resume_branch,
            mode,
            batch_name,
            get_proposal_description=(
                lambda df, ep: description),  # type: ignore
            get_proposal_commit_message=lambda ep: commit_message,
            get_proposal_title=lambda ep: title,
            allow_create_proposal=True,
            forge=forge,
            labels=labels,
            overwrite_existing=overwrite,
            derived_owner=derived_owner,
            existing_proposal=existing_proposal,
        )
    except UnsupportedForge as e:
        logging.exception(
            "No known supported forge for %s. Run 'svp login'?",
            full_branch_url(e.branch),
        )
        raise
    except InsufficientChangesForNewProposal:
        logging.info('Insufficient changes for a new merge proposal')
        raise
    except DivergedBranches:
        if not resume_branch:
            raise UnrelatedBranchExists()
        logging.warning('Branch exists that has diverged')
        raise
    except ForgeLoginRequired as e:
        logging.exception(
            "Credentials for hosting site at %r missing. "
            "Run 'svp login'?",
            e.forge.base_url,
        )
        raise

    if publish_result.proposal:
        if publish_result.is_new:
            logging.info("Merge proposal created.")
        else:
            logging.info("Merge proposal updated.")
        if publish_result.proposal.url:
            logging.info("URL: %s", publish_result.proposal.url)
        logging.info(
            "Description: %s", publish_result.proposal.get_description())
    return publish_result


def publish(directory, *, selector=None):
    batch = load_batch_metadata(directory)
    try:
        batch_name = batch['name']
    except KeyError:
        logging.error('no name found in %s', directory)
        return 1
    work = batch.get('work', [])
    if not work:
        logging.error('no work found in %s', directory)
        return 0
    errors = 0
    try:
        for name in list(work):
            entry = work[name]
            if selector and not selector(name, entry):
                continue
            try:
                publish_result = publish_one(
                    entry['url'], os.path.join(directory, name), batch_name,
                    entry['mode'],
                    subpath=entry.get('subpath', ''),
                    labels=entry.get('labels', []),
                    derived_owner=entry.get('derived-owner'),
                    commit_message=entry.get('commit-message'),
                    title=entry.get('title'),
                    existing_proposal_url=entry.get('proposal-url'),
                    description=entry.get('description'))   # type: ignore
            except EmptyMergeProposal:
                logging.info('No changes left')
                del work[name]
            except UnrelatedBranchExists:
                errors += 1
            else:
                if publish_result.mode == 'push':
                    drop_batch_entry(directory, name)
                    del work[name]
                elif publish_result.proposal:
                    entry['proposal-url'] = publish_result.proposal.url
            save_batch_metadata(directory, batch)
    finally:
        save_batch_metadata(directory, batch)
    if not work:
        logging.info('No work left in batch.yaml; you can now remove %s',
                     directory)
    if errors:
        return 1
    return 0


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    import argparse
    parser = argparse.ArgumentParser("svp batch")
    subparsers = parser.add_subparsers(dest="command")
    generate_parser = subparsers.add_parser("generate")
    generate_parser.add_argument(
        "--recipe", type=str, help="Recipe to use.")
    generate_parser.add_argument(
        "--candidates", type=str, help="File with candidate list.")
    generate_parser.add_argument('directory', nargs='?')
    publish_parser = subparsers.add_parser("publish")
    publish_parser.add_argument('directory')
    publish_parser.add_argument('name', nargs='?')
    status_parser = subparsers.add_parser("status")
    status_parser.add_argument('directory')
    status_parser.add_argument('codebase', nargs='?', default=None)
    diff_parser = subparsers.add_parser("diff")
    diff_parser.add_argument('directory')
    diff_parser.add_argument('codebase')
    args = parser.parse_args(argv)
    if args.command == "generate":
        if args.recipe:
            recipe = Recipe.from_path(args.recipe)
        else:
            parser.error('no recipe specified')
        if args.candidates:
            candidates = CandidateList.from_path(args.candidates)
        else:
            parser.error('no candidate list specified')
        if args.directory is None:
            args.directory = recipe.name
            logging.info('Using output directory: %s', args.directory)
        generate(
            recipe, candidates, args.directory,
            recipe_path=os.path.relpath(args.recipe, args.directory))
        logging.info(
            'Now, review the patches under %s, edit %s/batch.yaml as '
            'appropriate and then run "svp batch publish %s"',
            args.directory, args.directory, args.directory)
    elif args.command == 'publish':
        selector: Optional[Callable]
        if args.name:
            def selector(n, e):
                return n == args.name
        else:
            selector = None
        publish(args.directory, selector=selector)
        logging.info(
            'To see the status of open merge requests, run: '
            '"svp batch status %s"', args.directory)
    elif args.command == 'status':
        status(args.directory, args.codebase)
    elif args.command == 'diff':
        diff(args.directory, args.codebase)
    else:
        parser.print_usage()
    return 0



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from typing import Dict, Optional

from breezy.branch import Branch
from breezy.workingtree import WorkingTree
from . import _svp_rs

create_temp_sprout = _svp_rs.create_temp_sprout
PreCheckFailed = _svp_rs.PreCheckFailed
PostCheckFailed = _svp_rs.PostCheckFailed
run_pre_check = _svp_rs.run_pre_check
run_post_check = _svp_rs.run_post_check


class TemporarySprout:
    """Create a temporary sprout of a branch.

    This attempts to fetch the least amount of history as possible.
    """

    def __init__(
        self,
        branch: Branch,
        additional_colocated_branches: Optional[Dict[str, str]] = None,
        dir: Optional[str] = None,
    ) -> None:
        self.branch = branch
        self.additional_colocated_branches = additional_colocated_branches
        self.dir = dir

    def __enter__(self) -> WorkingTree:
        tree, self._destroy = create_temp_sprout(
            self.branch,
            additional_colocated_branches=self.additional_colocated_branches,
            dir=self.dir,
        )
        return tree

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._destroy()
        return False


BranchTemporarilyUnavailable = _svp_rs.BranchTemporarilyUnavailable
BranchRateLimited = _svp_rs.BranchRateLimited
BranchUnavailable = _svp_rs.BranchUnavailable
BranchMissing = _svp_rs.BranchMissing
BranchUnsupported = _svp_rs.BranchUnsupported
open_branch = _svp_rs.open_branch
open_branch_containing = _svp_rs.open_branch_containing
full_branch_url = _svp_rs.full_branch_url


def get_branch_vcs_type(branch):
    vcs = getattr(branch.repository, "vcs", None)
    if vcs:
        return vcs.abbreviation
    else:
        return "bzr"



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy.tests import TestCaseWithTransport

from silver_platter.utils import (
    PostCheckFailed,
    PreCheckFailed,
    TemporarySprout,
    run_post_check,
    run_pre_check,
)


class TemporarySproutTests(TestCaseWithTransport):
    def test_simple(self):
        builder = self.make_branch_builder(".")
        builder.start_series()
        orig_revid = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("debian/", None, "directory", "")),
                ("add", ("debian/control", None, "file", b"initial")),
            ],
            message="Initial\n",
        )
        builder.finish_series()
        branch = builder.get_branch()
        with TemporarySprout(branch, dir=self.test_dir) as tree:
            self.assertNotEqual(branch.control_url, tree.branch.control_url)
            tree.commit("blah")
            # Commits in the temporary sprout don't affect the original branch.
            self.assertEqual(branch.last_revision(), orig_revid)

    def test_nonexistent_colocated(self):
        # Colocated branches that are specified but don't exist are ignored.
        builder = self.make_branch_builder(".")
        builder.start_series()
        orig_revid = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("debian/", None, "directory", "")),
            ],
            message="Initial\n",
        )
        builder.finish_series()
        branch = builder.get_branch()
        with TemporarySprout(
                branch, {"foo": "foo"}, dir=self.test_dir) as tree:
            self.assertNotEqual(branch.control_url, tree.branch.control_url)
            tree.commit("blah")
            # Commits in the temporary sprout don't affect the original branch.
            self.assertEqual(branch.last_revision(), orig_revid)


class RunPreCheckTests(TestCaseWithTransport):
    def test_false(self):
        tree = self.make_branch_and_tree("tree")
        self.assertRaises(PreCheckFailed, run_pre_check, tree, "/bin/false")

    def test_true(self):
        tree = self.make_branch_and_tree("tree")
        self.assertIsNone(run_pre_check(tree, "/bin/true"))


class RunPostCheckTests(TestCaseWithTransport):
    def test_false(self):
        tree = self.make_branch_and_tree("tree")
        cid = tree.commit("a")
        self.assertRaises(
            PostCheckFailed, run_post_check, tree, "/bin/false",
            since_revid=cid
        )

    def test_true(self):
        tree = self.make_branch_and_tree("tree")
        cid = tree.commit("a")
        self.assertIsNone(run_post_check(tree, "/bin/true", since_revid=cid))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Automatic proposal/push creation."""

import argparse
import logging
import subprocess
import sys
from typing import List, Optional

import silver_platter  # noqa: F401

from .apply import (
    ScriptFailed,
    ScriptMadeNoChanges,
    ScriptNotFound,
    script_runner,
)
from .proposal import (
    ForgeLoginRequired,
    MergeProposal,
    UnsupportedForge,
    enable_tag_pushing,
    find_existing_proposed,
    get_forge,
)
from .publish import SUPPORTED_MODES, InsufficientChangesForNewProposal
from .utils import (
    BranchMissing,
    BranchUnavailable,
    BranchUnsupported,
    full_branch_url,
    open_branch,
)
from .workspace import Workspace

from ._svp_rs import derived_branch_name


def apply_and_publish(  # noqa: C901
        url: str, name: str, command: str, mode: str,
        commit_pending: Optional[bool] = None,
        labels: Optional[List[str]] = None, diff: bool = False,
        verify_command: Optional[str] = None,
        derived_owner: Optional[str] = None,
        refresh: bool = False, allow_create_proposal=None,
        get_commit_message=None, get_title=None, get_description=None):
    try:
        main_branch = open_branch(url)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        logging.exception("%s: %s", url, e)
        return 2

    overwrite = False

    try:
        forge = get_forge(main_branch)
    except UnsupportedForge as e:
        if mode != "push":
            raise
        # We can't figure out what branch to resume from when there's no forge
        # that can tell us.
        resume_branch = None
        existing_proposals: Optional[List[MergeProposal]] = []
        logging.warn(
            "Unsupported forge (%s), will attempt to push to %s",
            e,
            full_branch_url(main_branch),
        )
    else:
        (resume_branch, resume_overwrite,
         existing_proposals) = find_existing_proposed(
            main_branch, forge, name, owner=derived_owner
        )
        if resume_overwrite is not None:
            overwrite = resume_overwrite
    if refresh:
        if resume_branch:
            overwrite = True
        resume_branch = None

    if existing_proposals and len(existing_proposals) > 1:
        logging.warning(
            'Multiple open merge proposals for branch at %s: %r',
            resume_branch.user_url,  # type: ignore
            [mp.url for mp in existing_proposals])
        existing_proposal = existing_proposals[0]
        logging.info('Updating %s', existing_proposal.url)
    else:
        existing_proposal = None

    with Workspace(main_branch, resume_branch=resume_branch) as ws:
        try:
            result = script_runner(ws.local_tree, command, commit_pending)
        except ScriptMadeNoChanges:
            logging.error("Script did not make any changes.")
            return 0
        except ScriptFailed:
            logging.error("Script failed to run.")
            return 2
        except ScriptNotFound:
            logging.error("Script could not be found.")
            return 2

        if verify_command:
            try:
                subprocess.check_call(
                    verify_command, shell=True, cwd=ws.local_tree.abspath(".")
                )
            except subprocess.CalledProcessError:
                logging.error("Verify command failed.")
                return 2

        enable_tag_pushing(ws.local_tree.branch)

        try:
            publish_result = ws.publish_changes(
                mode=mode,
                name=name,
                get_proposal_description=(
                    lambda df, ep: get_description(result, df, ep)),
                get_proposal_commit_message=(
                    lambda ep: get_commit_message(result, ep)),
                get_proposal_title=(
                    lambda ep: get_title(result, ep)),
                allow_create_proposal=(
                    lambda: allow_create_proposal(result)),
                forge=forge,
                labels=labels,
                overwrite_existing=overwrite,
                derived_owner=derived_owner,
                existing_proposal=existing_proposal,
            )
        except UnsupportedForge as e:
            logging.error(
                "No known supported forge for %s. Run 'svp login'?",
                full_branch_url(e.branch),
            )
            return 2
        except InsufficientChangesForNewProposal:
            logging.info('Insufficient changes for a new merge proposal')
            return 1
        except ForgeLoginRequired as e:
            logging.error(
                "Credentials for hosting site at %r missing. "
                "Run 'svp login'?",
                e.forge.base_url,
            )
            return 2

        if publish_result.proposal:
            if publish_result.is_new:
                logging.info("Merge proposal created.")
            else:
                logging.info("Merge proposal updated.")
            if publish_result.proposal.url:
                logging.info("URL: %s", publish_result.proposal.url)
            logging.info(
                "Description: %s", publish_result.proposal.get_description())

        if diff:
            ws.show_diff(sys.stdout.buffer)

        return 1


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "url", help="URL of branch to work on.", type=str, nargs="?")
    parser.add_argument(
        "--command", help="Path to script to run.", type=str)
    parser.add_argument(
        "--derived-owner", type=str, default=None,
        help="Owner for derived branches."
    )
    parser.add_argument(
        "--refresh",
        action="store_true",
        help="Refresh changes if branch already exists",
    )
    parser.add_argument(
        "--label", type=str, help="Label to attach", action="append",
        default=[]
    )
    parser.add_argument(
        "--name", type=str, help="Proposed branch name", default=None)
    parser.add_argument(
        "--diff", action="store_true", help="Show diff of generated changes."
    )
    parser.add_argument(
        "--mode",
        help="Mode for pushing",
        choices=SUPPORTED_MODES,
        default=None,
        type=str,
    )
    parser.add_argument(
        "--commit-pending",
        help="Commit pending changes after script.",
        choices=["yes", "no", "auto"],
        default=None,
        type=str,
    )
    parser.add_argument(
        "--verify-command", type=str, help="Command to run to verify changes."
    )
    parser.add_argument(
        "--recipe", type=str, help="Recipe to use.")
    parser.add_argument(
        "--candidates", type=str, help="File with candidate list.")
    args = parser.parse_args(argv)

    if args.recipe:
        from .recipe import Recipe
        recipe = Recipe.from_path(args.recipe)
    else:
        recipe = None

    if not args.url and not args.candidates:
        parser.error("url or candidates are required")

    urls = []

    if args.url:
        urls = [args.url]

    if args.candidates:
        from .candidates import CandidateList
        candidatelist = CandidateList.from_path(args.candidates)
        urls.extend([candidate.url for candidate in candidatelist])

    if args.commit_pending:
        commit_pending = {
            "auto": None, "yes": True, "no": False}[args.commit_pending]
    elif recipe:
        commit_pending = recipe.commit_pending
    else:
        commit_pending = None

    if args.command:
        command = args.command
    elif recipe.command:
        command = recipe.command
    else:
        logging.exception('No command specified.')
        return 1

    if args.name is not None:
        name = args.name
    elif recipe and recipe.name:
        name = recipe.name
    else:
        name = derived_branch_name(command)

    if args.mode:
        mode = args.mode
    elif recipe and recipe.mode:
        mode = recipe.mode
    else:
        mode = "recipe"

    refresh = args.refresh

    if recipe and not recipe.resume:
        refresh = True

    def allow_create_proposal(result):
        if result.value is None:
            return True
        if recipe.propose_threshold is not None:
            return result.value >= recipe.propose_threshold
        return True

    def get_commit_message(result, existing_proposal):
        if recipe:
            return recipe.render_merge_request_commit_message(result.context)
        if existing_proposal is not None:
            return existing_proposal.get_commit_message()
        return None

    def get_title(result, existing_proposal):
        if recipe:
            return recipe.render_merge_request_title(result.context)
        if existing_proposal is not None:
            return existing_proposal.get_title()
        return None

    def get_description(result, description_format, existing_proposal):
        if recipe:
            description = recipe.render_merge_request_description(
                description_format, result.context)
            if description:
                return description
        if result.description is not None:
            return result.description
        if existing_proposal is not None:
            return existing_proposal.get_description()
        raise ValueError("No description available")

    retcode = 0

    for url in urls:
        result = apply_and_publish(
                url, name=name, command=command, mode=mode,
                commit_pending=commit_pending,
                labels=args.label, diff=args.diff,
                derived_owner=args.derived_owner, refresh=refresh,
                allow_create_proposal=allow_create_proposal,
                get_commit_message=get_commit_message,
                get_title=get_title,
                get_description=get_description)
        retcode = max(retcode, result)

    return retcode


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij
#                    Filippo Giunchedi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy.tests import TestCaseWithTransport

from silver_platter.candidates import CandidateList


class TestReadCandidates(TestCaseWithTransport):

    def test_read(self):
        self.build_tree_contents([('candidates.yaml', """\
---
- url: https://foo
""")])
        candidates = CandidateList.from_path('candidates.yaml')
        self.assertEqual(len(candidates.candidates), 1)



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
from io import BytesIO

from breezy.tests import TestCaseWithTransport

from silver_platter.workspace import Workspace


class WorkspaceTests(TestCaseWithTransport):
    def test_simple(self):
        b = self.make_branch("target")
        with Workspace(b, dir=self.test_dir) as ws:
            self.assertIsNone(ws.resume_branch)
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit("foo")
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_main())

    def test_with_resume(self):
        b = self.make_branch_and_tree("target")
        c = b.controldir.sprout("resume").open_workingtree()
        c.commit("some change")
        with Workspace(
                b.branch, resume_branch=c.branch, dir=self.test_dir) as ws:
            self.assertEqual(
                ws.local_tree.branch.last_revision(), c.branch.last_revision()
            )
            self.assertIs(ws.resume_branch, c.branch)
            self.assertTrue(ws.changes_since_main())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit("foo")
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())

    def test_with_resume_conflicting(self):
        b = self.make_branch_and_tree("target")
        self.build_tree_contents([("target/foo", "somecontents\n")])
        b.add(["foo"])
        b.commit("initial")
        c = b.controldir.sprout("resume").open_workingtree()
        self.build_tree_contents([("target/foo", "new contents in main\n")])
        b.commit("add conflict in main")
        self.build_tree_contents([("resume/foo", "new contents in resume\n")])
        c.commit("add conflict in resume")
        with Workspace(
                b.branch, resume_branch=c.branch, dir=self.test_dir) as ws:
            self.assertTrue(ws.refreshed)
            self.assertEqual(ws.base_revid, b.branch.last_revision())
            self.assertEqual(
                b.branch.last_revision(), ws.local_tree.branch.last_revision()
            )
            self.assertFalse(ws.changes_since_main())
            self.assertFalse(ws.changes_since_base())
            ws.local_tree.commit("foo")
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())

    def test_base_tree(self):
        b = self.make_branch_and_tree("target")
        cid = b.commit("some change")
        with Workspace(b.branch, dir=self.test_dir) as ws:
            ws.local_tree.commit("blah")
            self.assertEqual(cid, ws.base_tree().get_revision_id())

    def test_show_diff(self):
        b = self.make_branch_and_tree("target")
        with Workspace(b.branch, dir=self.test_dir) as ws:
            self.build_tree_contents(
                [(os.path.join(ws.local_tree.basedir, "foo"),
                  "some content\n")]
            )
            ws.local_tree.add(["foo"])
            ws.local_tree.commit("blah")
            self.assertTrue(ws.changes_since_main())
            self.assertTrue(ws.changes_since_base())
            f = BytesIO()
            ws.show_diff(outf=f)
            self.assertContainsRe(
                f.getvalue().decode("utf-8"), "\\+some content")



#!/usr/bin/python
# Copyright (C) 2022 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy.tests import TestCaseWithTransport

from silver_platter.apply import ScriptMadeNoChanges, script_runner


class TestScriptRunner(TestCaseWithTransport):

    def test_no_api(self):
        tree = self.make_branch_and_tree('t')
        self.build_tree_contents([
            ('script.sh', """\
#!/bin/sh
echo foo > bar
echo Did a thing
"""),
            ('t/bar', 'initial')])
        os.chmod('script.sh', 0o755)
        tree.add('bar')
        old_revid = tree.commit('initial')
        result = script_runner(
            tree,
            script=os.path.abspath('script.sh'),
            committer='Joe Example <joe@example.com>')
        self.assertFalse(tree.has_changes())
        self.assertEqual(result.old_revision, old_revid)
        self.assertEqual(result.new_revision, tree.last_revision())
        self.assertEqual(result.description, 'Did a thing\n')

    def test_api(self):
        tree = self.make_branch_and_tree('t')
        self.build_tree_contents([
            ('script.sh', """\
#!/bin/sh
echo foo > bar
echo '{"description": "Did a thing", "code": "success"}' > $SVP_RESULT
"""),
            ('t/bar', 'initial')])
        os.chmod('script.sh', 0o755)
        tree.add('bar')
        old_revid = tree.commit('initial')
        result = script_runner(
            tree,
            script=os.path.abspath('script.sh'),
            committer='Joe Example <joe@example.com>')
        self.assertFalse(tree.has_changes())
        self.assertEqual(result.old_revision, old_revid)
        self.assertEqual(result.new_revision, tree.last_revision())
        self.assertEqual(result.description, 'Did a thing')

    def test_new_file(self):
        tree = self.make_branch_and_tree('t')
        self.build_tree_contents([
            ('script.sh', """\
#!/bin/sh
echo foo > bar
echo Did a thing
""")])
        os.chmod('script.sh', 0o755)
        old_revid = tree.commit('initial')
        result = script_runner(
            tree,
            script=os.path.abspath('script.sh'),
            committer='Joe Example <joe@example.com>')
        self.assertFalse(tree.has_changes())
        self.assertEqual(result.old_revision, old_revid)
        self.assertEqual(result.new_revision, tree.last_revision())
        self.assertEqual(result.description, 'Did a thing\n')

    def test_no_changes(self):
        tree = self.make_branch_and_tree('t')
        self.build_tree_contents([
            ('script.sh', """\
#!/bin/sh
echo Did a thing
"""),
            ('t/bar', 'initial')])
        os.chmod('script.sh', 0o755)
        tree.add('bar')
        tree.commit('initial')
        self.assertRaises(
            ScriptMadeNoChanges, script_runner,
            tree,
            script=os.path.abspath('script.sh'),
            committer='Joe Example <joe@example.com>')
